#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import traceback
from typing import Any, Dict, Iterable

from airbyte_cdk.models import AirbyteLogMessage, AirbyteMessage, Level
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.sources.file_based.file_based_stream_permissions_reader import (
    AbstractFileBasedStreamPermissionsReader,
)
from airbyte_cdk.sources.file_based.stream import DefaultFileBasedStream
from airbyte_cdk.sources.file_based.types import StreamSlice
from airbyte_cdk.sources.streams.core import JsonSchema
from airbyte_cdk.sources.utils.record_helper import stream_data_to_airbyte_message


class PermissionsFileBasedStream(DefaultFileBasedStream):
    """
    A specialized stream for handling file-based ACL permissions.

    This stream works with the stream_reader to:
    1. Fetch ACL permissions for each file in the source
    2. Transform permissions into a standardized format
    3. Generate records containing permission information

    The stream_reader is responsible for the actual implementation of permission retrieval
    and schema definition, while this class handles the streaming interface.
    """

    def __init__(
        self, stream_permissions_reader: AbstractFileBasedStreamPermissionsReader, **kwargs: Any
    ):
        super().__init__(**kwargs)
        self.stream_permissions_reader = stream_permissions_reader

    def _filter_schema_invalid_properties(
        self, configured_catalog_json_schema: Dict[str, Any]
    ) -> Dict[str, Any]:
        return self.stream_permissions_reader.file_permissions_schema

    def read_records_from_slice(self, stream_slice: StreamSlice) -> Iterable[AirbyteMessage]:
        """
        Yield permissions records from all remote files
        """

        for file in stream_slice["files"]:
            no_permissions = False
            file_datetime_string = file.last_modified.strftime(self.DATE_TIME_FORMAT)
            try:
                permissions_record = self.stream_permissions_reader.get_file_acl_permissions(
                    file, logger=self.logger
                )
                if not permissions_record:
                    no_permissions = True
                    self.logger.warning(
                        f"Unable to fetch permissions. stream={self.name} file={file.uri}"
                    )
                    continue
                permissions_record = self.transform_record(
                    permissions_record, file, file_datetime_string
                )
                yield stream_data_to_airbyte_message(
                    self.name, permissions_record, is_file_transfer_message=False
                )
            except Exception as e:
                self.logger.error(f"Failed to retrieve permissions for file {file.uri}: {str(e)}")
                yield AirbyteMessage(
                    type=MessageType.LOG,
                    log=AirbyteLogMessage(
                        level=Level.ERROR,
                        message=f"Error retrieving files permissions: stream={self.name} file={file.uri}",
                        stack_trace=traceback.format_exc(),
                    ),
                )
            finally:
                if no_permissions:
                    yield AirbyteMessage(
                        type=MessageType.LOG,
                        log=AirbyteLogMessage(
                            level=Level.WARN,
                            message=f"Unable to fetch permissions. stream={self.name} file={file.uri}",
                        ),
                    )

    def _get_raw_json_schema(self) -> JsonSchema:
        """
        Retrieve the raw JSON schema for file permissions from the stream reader.

        Returns:
           The file permissions schema that defines the structure of permission records
        """
        return self.stream_permissions_reader.file_permissions_schema



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import Any, Callable, MutableMapping, Optional, Tuple

from airbyte_cdk.sources.streams.concurrent.cursor import CursorField
from airbyte_cdk.sources.streams.concurrent.state_converters.abstract_stream_state_converter import (
    AbstractStreamStateConverter,
    ConcurrencyCompatibleStateType,
)


class IncrementingCountStreamStateConverter(AbstractStreamStateConverter):
    def _from_state_message(self, value: Any) -> Any:
        return value

    def _to_state_message(self, value: Any) -> Any:
        return value

    @classmethod
    def get_end_provider(cls) -> Callable[[], float]:
        return lambda: float("inf")

    def convert_from_sequential_state(
        self,
        cursor_field: "CursorField",  # to deprecate as it is only needed for sequential state
        stream_state: MutableMapping[str, Any],
        start: Optional[Any],
    ) -> Tuple[Any, MutableMapping[str, Any]]:
        """
        Convert the state message to the format required by the ConcurrentCursor.

        e.g.
        {
            "state_type": ConcurrencyCompatibleStateType.date_range.value,
            "metadata": { … },
            "slices": [
                {"start": "10", "end": "2021-01-18T21:18:20.000+00:00"},
            ]
        }
        """
        sync_start = self._get_sync_start(cursor_field, stream_state, start)
        if self.is_state_message_compatible(stream_state):
            return sync_start, stream_state

        # Create a slice to represent the records synced during prior syncs.
        # The start and end are the same to avoid confusion as to whether the records for this slice
        # were actually synced
        slices = [
            {
                self.START_KEY: start if start is not None else sync_start,
                self.END_KEY: sync_start,  # this may not be relevant anymore
                self.MOST_RECENT_RECORD_KEY: sync_start,
            }
        ]

        return sync_start, {
            "state_type": ConcurrencyCompatibleStateType.integer.value,
            "slices": slices,
            "legacy": stream_state,
        }

    def parse_value(self, value: int) -> int:
        return value

    @property
    def zero_value(self) -> int:
        return 0

    def increment(self, value: int) -> int:
        return value + 1

    def output_format(self, value: int) -> int:
        return value

    def _get_sync_start(
        self,
        cursor_field: CursorField,
        stream_state: MutableMapping[str, Any],
        start: Optional[int],
    ) -> int:
        sync_start = start if start is not None else self.zero_value
        prev_sync_low_water_mark: Optional[int] = (
            stream_state[cursor_field.cursor_field_key]
            if cursor_field.cursor_field_key in stream_state
            else None
        )
        if prev_sync_low_water_mark and prev_sync_low_water_mark >= sync_start:
            return prev_sync_low_water_mark
        else:
            return sync_start



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import ABC

from airbyte_cdk.sources.declarative.requesters.request_options.request_options_provider import (
    RequestOptionsProvider,
)
from airbyte_cdk.sources.streams.concurrent.partitions.stream_slicer import (
    StreamSlicer as ConcurrentStreamSlicer,
)


class StreamSlicer(ConcurrentStreamSlicer, RequestOptionsProvider, ABC):
    """
    Slices the stream into a subset of records.
    Slices enable state checkpointing and data retrieval parallelization.

    The stream slicer keeps track of the cursor state as a dict of cursor_field -> cursor_value

    See the stream slicing section of the docs for more information.
    """

    pass



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

import json
from typing import Any, Mapping


class PerPartitionKeySerializer:
    """
    We are concerned of the performance of looping through the `states` list and evaluating equality on the partition. To reduce this
    concern, we wanted to use dictionaries to map `partition -> cursor`. However, partitions are dict and dict can't be used as dict keys
    since they are not hashable. By creating json string using the dict, we can have a use the dict as a key to the dict since strings are
    hashable.
    """

    @staticmethod
    def to_partition_key(to_serialize: Any) -> str:
        # separators have changed in Python 3.4. To avoid being impacted by further change, we explicitly specify our own value
        return json.dumps(to_serialize, indent=None, separators=(",", ":"), sort_keys=True)

    @staticmethod
    def to_partition(to_deserialize: Any) -> Mapping[str, Any]:
        return json.loads(to_deserialize)  # type: ignore # The partition is known to be a dict, but the type hint is Any



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from airbyte_cdk.sources.declarative.requesters.http_requester import HttpRequester
from airbyte_cdk.sources.declarative.requesters.request_option import RequestOption
from airbyte_cdk.sources.declarative.requesters.requester import Requester

__all__ = ["HttpRequester", "RequestOption", "Requester"]



#
# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
#
import logging
import os
from typing import Any, Dict, Iterable

from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader
from airbyte_cdk.sources.file_based.remote_file import RemoteFile

AIRBYTE_STAGING_DIRECTORY = os.getenv("AIRBYTE_STAGING_DIRECTORY", "/staging/files")
DEFAULT_LOCAL_DIRECTORY = "/tmp/airbyte-file-transfer"


class FileTransfer:
    def __init__(self) -> None:
        self._local_directory = (
            AIRBYTE_STAGING_DIRECTORY
            if os.path.exists(AIRBYTE_STAGING_DIRECTORY)
            else DEFAULT_LOCAL_DIRECTORY
        )

    def get_file(
        self,
        config: FileBasedStreamConfig,
        file: RemoteFile,
        stream_reader: AbstractFileBasedStreamReader,
        logger: logging.Logger,
    ) -> Iterable[Dict[str, Any]]:
        try:
            yield stream_reader.get_file(
                file=file, local_directory=self._local_directory, logger=logger
            )
        except Exception as ex:
            logger.error("An error has occurred while getting file: %s", str(ex))
            raise ex



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""Hashing utils for Airbyte."""

from __future__ import annotations

import hashlib
from collections.abc import Mapping

HASH_SEED = "Airbyte:"
"""Additional seed for randomizing one-way hashed strings."""


def one_way_hash(
    obj: Mapping[str, str] | list[str] | object,
    /,
) -> str:
    """Return a one-way hash of the given string.

    To ensure a unique domain of hashes, we prepend a seed to the string before hashing.
    """
    string_to_hash: str
    if isinstance(obj, Mapping):
        # Recursively sort and convert nested dictionaries to tuples of key-value pairs
        string_to_hash = str(sorted((k, one_way_hash(v)) for k, v in obj.items()))

    elif isinstance(obj, list):
        # Recursively hash elements of the list
        string_to_hash = str([one_way_hash(item) for item in obj])

    else:
        # Convert the object to a string
        string_to_hash = str(obj)

    return hashlib.sha256((HASH_SEED + str(string_to_hash)).encode()).hexdigest()



from .config.abstract_file_based_spec import AbstractFileBasedSpec
from .config.csv_format import CsvFormat
from .config.file_based_stream_config import FileBasedStreamConfig
from .config.jsonl_format import JsonlFormat
from .exceptions import CustomFileBasedException, ErrorListingFiles, FileBasedSourceError
from .file_based_source import DEFAULT_CONCURRENCY, FileBasedSource
from .file_based_stream_reader import AbstractFileBasedStreamReader, FileReadMode
from .remote_file import RemoteFile
from .stream.cursor import DefaultFileBasedCursor

__all__ = [
    "AbstractFileBasedSpec",
    "AbstractFileBasedStreamReader",
    "CsvFormat",
    "CustomFileBasedException",
    "DefaultFileBasedCursor",
    "ErrorListingFiles",
    "FileBasedSource",
    "FileBasedSourceError",
    "FileBasedStreamConfig",
    "FileReadMode",
    "JsonlFormat",
    "RemoteFile",
]
